// Copyright 2022 Esri.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import UIKit
import ArcGIS

class ViewController: UIViewController,
                        UIPickerViewDelegate,
                        UIPickerViewDataSource,
                        AGSGeoViewTouchDelegate {


    private typealias Category = (title: String, color: UIColor)
    private let categories: [Category] = [("Coffee shop", .brown),
                                          ("Gas station", .orange),
                                          ("Food", .cyan),
                                          ("Hotel", .blue),
                                          ("Neighborhood", .black),
                                          ("Parks and Outdoors", .green)]


    private let locatorTask = AGSLocatorTask(url: URL(string: "https://geocode-api.arcgis.com/arcgis/rest/services/World/GeocodeServer")!)
    private var cancelableGeocodeTask: AGSCancelable?

    private let graphicsOverlay = AGSGraphicsOverlay()
    private var isNavigatingObserver: NSKeyValueObservation?
    
    private struct AttributeKeys {
        static let placeAddress = "Place_addr"
        static let placeName = "PlaceName"
    }

    
    @IBOutlet weak var mapView: AGSMapView!

    @IBOutlet weak var categoryPicker: UIPickerView!

    
    // Perform one-time initialization of the map and events.

    private func setupMap() {

        let map = AGSMap(
            basemapStyle: .arcGISNavigation
        )
        mapView.map = map
        mapView.setViewpoint(
            AGSViewpoint(
                latitude: 34.09042,
                longitude: -118.71511,
                scale: 500_000
            )
        )

        mapView.touchDelegate = self
        mapView.graphicsOverlays.add(graphicsOverlay)

        // When the map finishes navigating to a new extent perform a new search using the updated visible area.
        isNavigatingObserver = mapView.observe(\.isNavigating, options:[]) { (mapView, _) in

            guard !mapView.isNavigating else { return }

            // Update results if the map view extent has moved.
            DispatchQueue.main.async { [weak self] in
                self?.findPlacesForCategoryPickerSelection()
            }
        }

        // After the view loads, perform a search on the selected item in the picker. Once this completes
        // remove this handler.
        mapView.viewpointChangedHandler = { [weak self] () -> Void in
            DispatchQueue.main.async {
                self?.findPlacesForCategoryPickerSelection()
                self?.mapView.viewpointChangedHandler = nil
            }
        }

    }

    

    // Get the current selection in the picker and find places for that category.
    private func findPlacesForCategoryPickerSelection() {
        let categoryIndex = categoryPicker.selectedRow(inComponent: 0)
        guard categoryIndex < categories.count else { return }
        let category = categories[categoryIndex]
        findPlaces(forCategory: category)
    }

    
    // When the user selects a category, run a geocode task and find nearby places at the center of the current visible extent.

    private func findPlaces(forCategory category: Category) {
        
        guard let visibleArea = mapView.visibleArea else { return }

        // Clean up anything from the prior search.
        mapView.callout.dismiss()
        graphicsOverlay.graphics.removeAllObjects()
        cancelableGeocodeTask?.cancel()

        // Configure parameters for the search task.
        let geocodeParameters = AGSGeocodeParameters()
        geocodeParameters.preferredSearchLocation = visibleArea.extent.center
        geocodeParameters.maxResults = 25
        geocodeParameters.resultAttributeNames.append(contentsOf: [AttributeKeys.placeAddress, AttributeKeys.placeName])

        cancelableGeocodeTask = locatorTask.geocode(withSearchText: category.title, parameters: geocodeParameters) { [weak self] (results: [AGSGeocodeResult]?, error: Error?) -> Void in
            
            guard let self = self else { return }
            guard error == nil else {
                print("geocode error", error!.localizedDescription)
                return
            }
            guard let results = results, results.count > 0 else {
                print("No places found for category", category.title)
                return
            }
            
            // Represent each located place as a dot on the map. Each graphic symbol also gets the place attributes so we can
            // show them later when the user taps the graphic.
            for result in results {
                let placeSymbol = AGSSimpleMarkerSymbol(style: .circle, color: category.color, size: 10.0)
                placeSymbol.outline = AGSSimpleLineSymbol(style: .solid, color: .white, width: 2)
                let graphic = AGSGraphic(geometry: result.displayLocation, symbol: placeSymbol, attributes: result.attributes as [String : AnyObject]?)
                self.graphicsOverlay.graphics.add(graphic)
            }
        }

    }

    

    // When a graphic in the graphics layer is tapped, use a callout to dislay its attributes.
    private func showCalloutForGraphic(_ graphic:AGSGraphic, tapLocation:AGSPoint) {
        self.mapView.callout.title = graphic.attributes["PlaceName"] as? String ?? "Unknown"
        self.mapView.callout.detail = graphic.attributes["Place_addr"] as? String ?? "no address provided"
        self.mapView.callout.isAccessoryButtonHidden = true
        self.mapView.callout.show(for: graphic, tapLocation: tapLocation, animated: true)
    }

    
    override func viewDidLoad() {
        super.viewDidLoad()
        setupMap()
    }


    func numberOfComponents(in pickerView: UIPickerView) -> Int {
        return 1
    }
    
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int {
        return categories.count
    }

    func pickerView(_ pickerView: UIPickerView, titleForRow row: Int, forComponent component: Int) -> String? {
        return categories[row].title
    }
    
    func pickerView(_ pickerView: UIPickerView, didSelectRow row: Int, inComponent component: Int) {
        findPlaces(forCategory: categories[row])
    }

    
    // When the user taps the map identify the graphic element(s) at that location. It is possible identify
    // can return more than one result, but we only show the first one found.

    func geoView(_ geoView: AGSGeoView, didTapAtScreenPoint screenPoint: CGPoint, mapPoint: AGSPoint) {
        
        // Dismiss the callout if already visible.
        self.mapView.callout.dismiss()
        
        // Identify graphics at the tapped location.
        self.mapView.identify(self.graphicsOverlay, screenPoint: screenPoint, tolerance: 10, returnPopupsOnly: false, maximumResults: 2) { (result: AGSIdentifyGraphicsOverlayResult) -> Void in
            guard result.error == nil else {
                print(result.error!)
                return
            }
            if let graphic = result.graphics.first {
                // Show a callout for the first graphic in the array.
                self.showCalloutForGraphic(graphic, tapLocation: mapPoint)
            }
        }
    }

}


